import inspect
import sys

# This script generates zroya stub file from documentation zroya C/C++ code.
# Stub file is a hint for IDE. It contains a list of all functions and methods in zroya with docstrings.

# TODO: Find better way for output formatting than this


def GenerateStubFile(path_to_pyd, path_to_python_files):

    # Add Path to .pyd file with zroya to import path
    sys.path.append(path_to_pyd)
    sys.path.append(path_to_python_files)

    # Import zroya module
    import _zroya
    import template_enums
    import dismiss_reason

    modules = [
        _zroya,
        template_enums,
        dismiss_reason
    ]

    # Generate .pyi file for zroya module
    with open("./zroya/zroya.pyi", "w", encoding="utf8") as output:

        for module in modules:

            # Get all classes/functions in zroya package
            for name, obj in inspect.getmembers(module):

                # TODO: Do not expose Enum class at all
                if "Enum" in name:
                    continue

                # Is it a package function
                if inspect.isbuiltin(obj):

                    # Print def <function_name>(<function_parameters>):
                    output.write("def {}{}:\n".format(name, inspect.signature(obj)))
                    # Print functions docstring
                    output.write("\t\"\"\"\n\t{}\n\t\"\"\"\n\tpass\n\n".format(inspect.getdoc(obj).replace("\n", "\n\t")))

                # Is it a class
                if inspect.isclass(obj):

                    # Print class <class_name>:
                    output.write("class {}:\n".format(name))
                    # Print class docstring
                    output.write("\t\"\"\"{}\"\"\"\n\n".format(obj.__doc__))

                    # Get all attributes of class
                    for member_name, member_value in inspect.getmembers(obj):

                        # Ignore zroya members
                        if member_name[0] != "_":
                            # Is it zroya enum?
                            if hasattr(member_value, "value"):
                                member_value = member_value.value

                            # Is value string
                            if isinstance( member_value, str):
                                # Put " and " around value
                                output.write("\t{} = \"{}\"\n\n".format(member_name, member_value))
                            elif isinstance( member_value, int):
                                # Print number
                                output.write("\t{} = {}\n\n".format(member_name, member_value))

                    # Get all methods from class
                    for cname, cobj in inspect.getmembers(obj):
                        # Is it a method? Ignoring those starting with undercore, because zroya does not set them.
                        if inspect.ismethoddescriptor(cobj) and cname[0] != "_":
                            # Get method signature. Replace leading and trailing bracket, because I have to add self parameter

                            signature = str(inspect.signature(cobj)).replace("(", "").replace(")", "")
                            # Print: def <function_name>(self, <function_parameters>):

                            if len(signature) == 0:
                                output.write("\tdef {}(self):\n".format(cname))
                            else:
                                output.write("\tdef {}(self, {}):\n".format(cname, signature))

                            # Print methods docstring
                            output.write("\t\t\"\"\"\n\t\t{}\n\t\t\"\"\"\n\t\tpass\n\n".format(inspect.getdoc(cobj).replace("\n", "\n\t\t")))


from enum import IntEnum, Enum


class AudioMode(IntEnum):
    """
    AudioMode is enumeration holds all valid parameters accepted by :py:meth:`zroya.Template.setAudio` method's `mode`
    parameter.

    Example:
        .. code-block:: python

            # t is an instance of zroya.Template
            t.setAudio( mode=zroya.AudioMode.Silence )

    """

    Default = 0
    """
    Selected audio will be played only once.
    """

    Silence = 1
    """
    No audio is played at all.
    """

    Loop = 2
    """
    Play audio in loop until it is moved to Action Center. This time may vary due to different user
    configuration.
    """


class Audio(Enum):
    """
    Audio enumeration contains values for  accepted values for `audio` parameter of :py:meth:`zroya.Template.setAudio`
    method.

    Example:
        .. code-block:: python

            # t is an instance of zroya.Template
            t.setAudio( audio=zroya.Audio.IM )

    """

    Default = "ms-winsoundevent:Notification.Default"
    """
    .. raw:: html
        
        <iframe src="_static/audio.html?file=Default"></iframe>

    """

    IM = "ms-winsoundevent:Notification.IM"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=IM"></iframe>

    """

    Mail = "ms-winsoundevent:Notification.Mail"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Email"></iframe>

    """

    Reminder = "ms-winsoundevent:Notification.Reminder"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Reminder"></iframe>

    """

    SMS = "ms-winsoundevent:Notification.SMS"
    Alarm = "ms-winsoundevent:Notification.Looping.Alarm"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm01&loop"></iframe>

    """

    Alarm2 = "ms-winsoundevent:Notification.Looping.Alarm2"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm02&loop"></iframe>

    """

    Alarm3 = "ms-winsoundevent:Notification.Looping.Alarm3"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm03&loop"></iframe>

    """

    Alarm4 = "ms-winsoundevent:Notification.Looping.Alarm4"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm04&loop"></iframe>

    """

    Alarm5 = "ms-winsoundevent:Notification.Looping.Alarm5"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm05&loop"></iframe>

    """

    Alarm6 = "ms-winsoundevent:Notification.Looping.Alarm6"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm06&loop"></iframe>

    """

    Alarm7 = "ms-winsoundevent:Notification.Looping.Alarm7"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm07&loop"></iframe>

    """

    Alarm8 = "ms-winsoundevent:Notification.Looping.Alarm8"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm08&loop"></iframe>

    """

    Alarm9 = "ms-winsoundevent:Notification.Looping.Alarm9"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm09&loop"></iframe>

    """

    Alarm10 = "ms-winsoundevent:Notification.Looping.Alarm10"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Alarm10&loop"></iframe>

    """

    Call = "ms-winsoundevent:Notification.Looping.Call"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring01&loop"></iframe>

    """

    Call2 = "ms-winsoundevent:Notification.Looping.Call2"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring02&loop"></iframe>

    """

    Call3 = "ms-winsoundevent:Notification.Looping.Call3"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring03&loop"></iframe>

    """

    Call4 = "ms-winsoundevent:Notification.Looping.Call4"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring04&loop"></iframe>

    """

    Call5 = "ms-winsoundevent:Notification.Looping.Call5"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring05&loop"></iframe>

    """

    Call6 = "ms-winsoundevent:Notification.Looping.Call6"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring06&loop"></iframe>

    """

    Call7 = "ms-winsoundevent:Notification.Looping.Call7"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring07&loop"></iframe>

    """

    Call8 = "ms-winsoundevent:Notification.Looping.Call8"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring08&loop"></iframe>

    """

    Call9 = "ms-winsoundevent:Notification.Looping.Call9"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring09&loop"></iframe>

    """

    Call10 = "ms-winsoundevent:Notification.Looping.Call10"
    """
    .. raw:: html

        <iframe src="_static/audio.html?file=Ring10&loop"></iframe>

    """


class TemplateType(IntEnum):
    """
    All possible values for :py:class:`zroya.Template` constructor.

    Example:
        .. code-block:: python

            zroya.Template(zroya.TemplateType.ImageAndText2)

    """

    ImageAndText1 = 0
    """
    A large image and a single string wrapped across three lines of text.
    """

    ImageAndText2 = 1
    """
    A large image, one string of bold text on the first line, one string of regular text wrapped across the
    second and third lines.
    """

    ImageAndText3 = 2
    """
    A large image, one string of bold text wrapped across the first two lines, one string of regular text on
    the third line.
    """

    ImageAndText4 = 3
    """
    A large image, one string of bold text on the first line, one string of regular text on the second line,
    one string of regular text on the third line.
    """

    Text1 = 4
    """
    Single string wrapped across three lines of text.
    """

    Text2 = 5
    """
    One string of bold text on the first line, one string of regular text wrapped across the second and third
    lines.
    """

    Text3 = 6
    """
    One string of bold text wrapped across the first two lines, one string of regular text on the third line.
    """

    Text4 = 7
    """
    One string of bold text on the first line, one string of regular text on the second line, one string of
    regular text on the third line.
    """



# -*- coding: utf-8 -*-

from shutil import rmtree
from setuptools import setup, Extension, Command
from setuptools.command.test import test
from setuptools.command.build_ext import build_ext
import generate_stubs
import os
import sys

sys.path.append(os.path.abspath("./zroya/"))

import version

here = os.path.abspath(os.path.dirname(__file__))

# Include path for _zroya module
includes_list = ["./module"]

# List of all *.cpp files in ./module directory
sources_list = []
for root, dirs, files in os.walk("./module"):
    for f in files:
        if os.path.splitext(f)[1] == ".cpp":
            sources_list.append(os.path.join(root, f))

# Python C/CPP Api extension configuration
ext_modules = [
    Extension("_zroya",
        sources=sources_list,
        include_dirs=includes_list
    )
]


def discover_and_run_tests():
    import unittest

    # use the default shared TestLoader instance
    test_loader = unittest.defaultTestLoader

    # use the basic test runner that outputs to sys.stderr
    test_runner = unittest.TextTestRunner()

    # automatically discover all tests
    # NOTE: only works for python 2.7 and later
    test_suite = test_loader.discover( os.path.join(here, "tests"))

    # run the test suite
    test_runner.run(test_suite)


def find_pyd_file():
    """
    Return path to .pyd after successful build command.
    :return: Path to .pyd file or None.
    """

    if not os.path.isdir("./build"):
        raise NotADirectoryError

    for path, dirs, files in os.walk("./build"):
        for file_name in files:
            file_name_parts = os.path.splitext(file_name)
            if file_name_parts[1] == ".pyd":
                return path
    return None


class StubsCommand(build_ext):

    description = "Generate python stubs with documentation from C code."

    def run(self):
        build_ext.run(self)

        print("running stubs")
        # Generate .pyd file for this module
        generate_stubs.GenerateStubFile(find_pyd_file(), os.path.abspath("./zroya/"))


class DocumentationCommand(Command):

    description = "Generate documentation from /docs_source directory."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system("{} {}".format(os.path.join(os.path.abspath("./docs_source/sphinx/"), "make.bat"), "html"))
        sys.exit()


class UploadCommand(Command):

    description = 'Build and publish the package.'
    user_options = [
        (
            "pypi=",
            "p",
            "Set PyPi repository in which should be distribution uploaded. Valid values are 'test' for TestPyPi or \
            'standard' for regular PyPi."
        )
    ]

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        self.pypi = "standard"

    def finalize_options(self):
        assert(self.pypi == "standard" or self.pypi == "test")

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')

        if self.pypi == "standard":
            os.system('twine upload --repository pypi dist/*')
        else:
            os.system('twine upload --repository testpypi dist/*')

        sys.exit()


class DiscoverTest(test):
    """
    Discover and run tests.
    See: https://stackoverflow.com/questions/17001010/how-to-run-unittest-discover-from-python-setup-py-test
    """

    description = "Run tests in /tests folder with unittest."

    def finalize_options(self):
        test.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        discover_and_run_tests()


setup(name='zroya',
    version=version.__version__,
    description='Python library for creating native Windows notifications.',
    long_description=open("README.md").read(),

    author='Jan Malčák',
    author_email='looorin@gmail.com',

    license='MIT',
    url='https://malja.github.io/zroya',

    keywords=["notifications", "windows", "toast"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Win32 (MS Windows)",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: C",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: User Interfaces"
    ],

    ext_modules=ext_modules,
    test_suite="tests",
    packages=["zroya"],
    cmdclass={
        "stubs": StubsCommand,
        'upload': UploadCommand,
        "docs": DocumentationCommand,
        "test": DiscoverTest,
    }
)



class DismissReason(object):

    """
    This class represents a notification dismiss reason. It is passed to callback registered in on_dismiss parameter
    of :py:func:`zroya.show` function.

    You can print it to get a reason description or compare it with any of following attributes.
    """

    User = 0
    """
    The user dismissed the toast.
    """

    App = 1
    """
    The application hid the toast using :py:func:`zroya.hide`.
    """

    Expired = 2
    """
    The toast has expired.
    """

    def __init__(self, reason):
        """
        Create an instance of dismiss reason. Zroya uses this class to return a dismiss reason of notification back to
        zroya callback. For instance, when you create a notification with on_dismiss callback set:

        .. code-block:: python

            def myDismissCallback(notificationID, reason):
                print("Notification {} was dismissed by user. Reason: {}.".format(notificationID, reason))

            # t is an instance of zroya.Template
            zroya.show(t, on_dismiss = myCallback)

        this class will be used as a second parameter 'reason'. Since this is kind of C->Python bridge class, you
        will probably never create an instance of it.

        Args:
            reason (int): Integer representation of C dismiss reason.
        """

        self._reason = reason

    def __str__(self):

        # Make sure, numbers corresponds with actual dismiss reasons from Windows core:
        # Search ToastDismissalReason::ToastDismissalReason_UserCanceled in wintoastlib.h

        if self._reason == DismissReason.User:
            return "The user dismissed the toast."
        if self._reason == DismissReason.App:
            return "The application hid the toast using zroya.hide."
        if self._reason == DismissReason.Expired:
            return "The toast has expired."

        return "Unknown dismiss reason. If you are seeing this, please report it as a bug. Thank you."

    def __eq__(self, other):
        if isinstance(other, int):
            return other == self._reason

        return False

    def __ne__(self, other):
        return not other == self
